//2307 see stl3.cpp
#include <iostream>
#include <string>
#include <thread> //mutexx
#include <chrono>


#include <mutex>
#include <condition_variable>
#include <atomic>
#include <shared_mutex>
#include <future>

using std::cout;
using std::endl;
using std::string;
using std::thread;

using std::mutex;
using std::timed_mutex;
using std::shared_mutex;
using std::lock_guard;
using std::unique_lock;
using std::try_to_lock;
using std::defer_lock;
using std::atomic;
using std::condition_variable;
using std::promise;
using std::future;
using std::shared_future;
using std::async;
using std::packaged_task;
using namespace std::chrono_literals; //100ms

using namespace std::chrono;

//condition_variable
void condition_variableThread(condition_variable &condVar, mutex &myMutex) 
{ //passing , unique_lock<mutex>& my_unique_lock seems causing crash when lock. should be another lock using same mutex
	cout << "condition_variableThread entered, locking\n"; //no locks yet, output may mess up
	unique_lock<mutex> my_unique_lock(myMutex);
	cout << "condition_variableThread locked, sleep one sec" << endl;
	std::this_thread::sleep_for(std::chrono::seconds(1));
	cout << "condition_variableThread locked, will wait, which unblock condition_variableTest" << endl;
	condVar.wait(my_unique_lock); //unlocks my_unique_lock, and wait for it unlocked by provider if provider locks.   // last param , [] {return dataReady; }
	bool bLockSuccessful = my_unique_lock.owns_lock();
	_ASSERT(bLockSuccessful);
	cout << "condition_variableThread wait done, doing consume work, bLockSuccessful = " << bLockSuccessful << endl;
	cout << "condition_variableThread exiting." << endl;
}
void condition_variableTest()
{
	mutex myMutex;
	condition_variable condVar;
	unique_lock<mutex> my_unique_lock(myMutex, defer_lock);
	thread t1(condition_variableThread, ref(condVar), ref(myMutex));
	cout << "condition_variableTest: condition_variableThread started, sleep one sec for thread to lock first" << endl;
	std::this_thread::sleep_for(std::chrono::seconds(1));
	cout << "condition_variableTest: will lock, this should wait since thread lock first, wait() unlocks it" << endl;
	my_unique_lock.lock();
	std::this_thread::sleep_for(std::chrono::seconds(1)); //to avoid both write to cout?
	cout << "condition_variableTest: got lock, doing producer work, will notify" << endl;
	std::this_thread::sleep_for(std::chrono::seconds(2));
	condVar.notify_one();
	cout << "condition_variableTest: notify_one done, wait a sec, then unlock\n";
	std::this_thread::sleep_for(std::chrono::seconds(1));
	my_unique_lock.unlock(); //required. notify_one is not enough to let wait() exit.
	cout << "condition_variableTest: unlocked, joining" << endl;
	t1.join();
	cout << "condition_variableTest: joined, exiting." << endl;
}

//240303 thread safe queue, see 3vr.c
#define WXC_THREAD_FUNC_STANALONE
//#define WXC_ERR_TEMPLATE_LAMBDA //not working
template<class T> class CQueueThread //not Deque for simplicity no both end, but add random access
{
private:
	mutable shared_mutex mMutex;
	deque<T> mDeque;
public:
	T at(size_t iIndex) { shared_lock lock(mMutex);	sleep_for(1s);	return mDeque.at(iIndex); }
	const T& front() const//requires mutable
	{
		shared_lock lock(mMutex);
		T value = mDeque.front();
		osyncstream(cout) << " fronting " << value << endl;
		sleep_for(1s);
		osyncstream(cout) << " front done " << value << endl;
		return value;
	}
	void push_back(const T& value);
	void pop_front()
	{
		unique_lock lock(mMutex);
		if (mDeque.empty())
			throw std::out_of_range("no more data");
		mDeque.pop_front();
		sleep_for(1s);
	}
	size_t size() { return mDeque.size(); }
};
template<class T> void CQueueThread<T>::push_back(const T& value)
{
	unique_lock lock(mMutex); //can also mMutex.lock(); mMutex.unlock();
	osyncstream(cout) << " pushing " << value << endl;
	mDeque.push_back(value);
	sleep_for(1s);
	osyncstream(cout) << " push " << value << " done" << endl;
}
template<class T> int funcThread1(CQueueThread<T>& Queue, int tag, int iAction)
{
	osyncstream(cout) << "thread tag = " << tag << (iAction == 0 ? " pushing": "reading ") << endl;
	if (iAction == 0)
	{
		Queue.push_back(22);
		return 0;
	}
	else
		return Queue.front();
}
void QueueThreadTest()
{ //output should be: fronting 11 fronting 11 without done in between since readonly . pushing 22	push 22 done together no 11 in between
	
	CQueueThread<int> Queue{};
	int iAction = 0;
#ifdef WXC_THREAD_FUNC_STANALONE
	future Future = async(funcThread1<int>, ref(Queue), 2, iAction);
#else
	#ifdef WXC_ERR_TEMPLATE_LAMBDA
	auto functionThread = []<class T>(CQueueThread<T>&Queue, int tag, int iAction)
	{
		funcThread1(Queue, tag, iAction);
};
	future<int> Future = async(functionThread<int>, ref(Queue), 2, iAction); //complain about <int>
	#else
	auto functionThread = [](CQueueThread<int>&Queue, int tag, int iAction)
	{
		return funcThread1(Queue, tag, iAction);
	};
	future<int> Future = async(functionThread, ref(Queue), 2, iAction);
	#endif
#endif

	cout << "main " << " push..." << endl;
	Queue.push_back(11);
	cout << "main " << ", push done" << endl;
	Future.wait();
	assert(Queue.size() == 2);
	//test readonly
	iAction = 1;
	Future = async(funcThread1<int>, ref(Queue), 2, iAction);
	cout << "main " << " read" << endl;
	int vMian = Queue.front();
	int vThread = Future.get();
	assert(vMian == vThread);
	//test pop
	while (Queue.size())
		Queue.pop_front();
	try{ Queue.pop_front(); }
	catch (std::out_of_range e){
		cout << "error empty queue, " << e.what() << endl;
	}
}

//240310 atomic mutex, see stl3.c
class busy_mutex
{
	atomic_flag mFlag;
public:
	busy_mutex() { mFlag.clear(); }
	void lock() { while (mFlag.test_and_set() ) yield(); }
	void unlock() { mFlag.clear(); };
};
void SyncWithAtomic(atomic_flag& flag, packaged_task<void()>&& task) //note && need work
{
	while (true)
	{
		if (!flag.test_and_set()) //if !flag flag = true
		{
			task();
			flag.clear(); //flag = false
			break;
		}
		sleep_for(10ms);
	}
}
void SyncWithAtomicTest()
{
	vector<int> aOutput;
	vector<int> aExpectAtomicYes = { 0,1, 10, 11 };
	vector<int> aExpectAtomicNo = { 10,0, 11, 1 };
	int iStartWith = 0;
	//sync with SyncWithAtomic()
	atomic_flag flag = ATOMIC_FLAG_INIT;
	auto lambda = [&aOutput, &iStartWith]()
		{
			int iStart = iStartWith;
			aOutput.push_back(iStart);
			iStart++;
			sleep_for(1s);
			aOutput.push_back(iStart);
			cout << "lambda" << endl; //better output
		};
	packaged_task<void()> task(lambda);
	SyncWithAtomic(flag, move(task));
	task.reset();
	iStartWith = 10;
	thread ThreadAtomicYes(SyncWithAtomic, ref(flag), move(task));
	ThreadAtomicYes.join();
	assert(aOutput == aExpectAtomicYes);
	//sync with busy_mutex
	busy_mutex Mutex;
	function funcNoCapture = [](int iStartWith, vector<int>& aOutput, busy_mutex & Mutex)
		{
			int iStart = iStartWith;
			Mutex.lock();
			aOutput.push_back(iStart);
			iStart++;
			sleep_for(1s);
			aOutput.push_back(iStart);
			cout << "funcNoCapture" << endl; //better output
			Mutex.unlock();
		};
	iStartWith = 0;
	aOutput.clear();
	funcNoCapture(iStartWith, aOutput, Mutex);
	iStartWith = 10;
	thread ThreadMutex = thread(funcNoCapture, iStartWith, ref(aOutput), ref(Mutex));
	ThreadMutex.join();
	assert(aOutput == aExpectAtomicYes);
	//no sync, therefore wrong result.
	aOutput.clear();
	packaged_task<void()> taskAtomicNo(lambda);
	iStartWith = 10;
	thread ThreadAtomicNo(lambda);
	sleep_for(1ms);
	iStartWith = 0;
	lambda();
	ThreadAtomicNo.join(); //diff result if breakpt here.
	assert(aOutput == aExpectAtomicNo);
}

